package Arithmetics;

import java.math.BigInteger;

/**
 * Created by Tzlil on 3/17/2016.
 */
public class Fp implements Arithmetic<BigInteger> {
    public final BigInteger p;

    public Fp(BigInteger p) {
        this.p = p;
    }

    @Override
    public BigInteger add(BigInteger a,BigInteger b){
        return a.add(b).mod(p);
    }

    @Override
    public BigInteger sub(BigInteger a,BigInteger b){
        return a.add(p).subtract(b).mod(p);
    }

    @Override
    public BigInteger mul(BigInteger a,BigInteger b){
        return a.multiply(b).mod(p);
    }

    @Override
    public BigInteger div(BigInteger a,BigInteger b){
        return mul(a,inv(b));
    }

    public BigInteger pow(BigInteger b,BigInteger e){
        if (e.compareTo(BigInteger.ZERO) < 0 ) {
            return pow(inv(b), e.negate());
        }
        BigInteger result = BigInteger.ONE;
        while (e.compareTo(BigInteger.ZERO) > 0) {
            if (e.testBit(0)) {
                result = mul(result, b);
            }
            e = e.shiftRight(1);
            b = mul(b, b);
        }
        return result;
    }

    public BigInteger inv(BigInteger a){
        return pow(a,p.subtract(BigInteger.valueOf(2)));
    }
}
